package ru.dbotthepony.kstarbound

import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.GLFW.glfwSetWindowShouldClose
import org.lwjgl.opengl.GL11.GL_LINES
import org.lwjgl.opengl.GL11.GL_RGBA
import org.lwjgl.opengl.GL46
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLFrameBuffer
import ru.dbotthepony.kstarbound.client.gl.GLTexture2D
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.client.gl.vertex.quad
import ru.dbotthepony.kstarbound.io.*
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.entities.PlayerEntity
import ru.dbotthepony.kvector.vector.Color
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.io.ByteArrayInputStream
import java.io.DataInputStream
import java.io.File
import java.util.zip.Inflater

private val LOGGER = LogManager.getLogger()

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	val db = BTreeDB(File("F:\\SteamLibrary\\steamapps\\common\\Starbound - Unstable\\storage\\universe\\389760395_938904237_-238610574_5.world"))
	//val db = BTreeDB(File("world.world"))

	val client = StarboundClient()

	//Starbound.addFilePath(File("./unpacked_assets/"))
	Starbound.addPakPath(File("J:\\Steam\\steamapps\\common\\Starbound\\assets\\packed.pak"))
	//Starbound.addPakPath(File("packed.pak"))

	/*Starbound.initializeGame { finished, replaceStatus, status ->
		client.putDebugLog(status, replaceStatus)
	}*/

	client.onTermination {
		Starbound.terminateLoading = true
	}

	val ent = PlayerEntity(client.world!!)

	Starbound.onInitialize {
		var find = 0L
		var set = 0L
		var parse = 0L

		for (chunkX in 0 .. 94) {
			for (chunkY in 0 .. 61) {
				var t = System.currentTimeMillis()
				val data = db.read(byteArrayOf(1, 0, chunkX.toByte(), 0, chunkY.toByte()))
				find += System.currentTimeMillis() - t

				if (data != null) {
					val chunk = client.world!!.computeIfAbsent(ChunkPos(chunkX, chunkY))
					val inflater = Inflater()
					inflater.setInput(data)

					t = System.currentTimeMillis()
					val output = ByteArray(64_000)
					val actual = inflater.inflate(output)
					val reader = DataInputStream(ByteArrayInputStream(output))
					parse += System.currentTimeMillis() - t

					reader.skipBytes(3)

					var hitTile = false

					t = System.currentTimeMillis()

					for (y in 0 .. 31) {
						for (x in 0 .. 31) {
							val materialID = reader.readUnsignedShort()
							val getMat = Starbound.tilesAccessID[materialID]

							if (getMat != null) {
								chunk.foreground[x, y].material = getMat
								hitTile = true
							}

							// reader.skipBytes(1) // Foreground hue shift
							// reader.skipBytes(1) // Foreground color variant
							val colorShift = reader.readUnsignedByte()
							val colorVariant = reader.readUnsignedByte()

							val modifier = reader.readUnsignedShort()
							val getModifier = Starbound.tileModifiersByIDAccess[modifier]

							chunk.foreground[x, y].color = colorVariant
							chunk.foreground[x, y].setHueShift(colorShift)

							if (getModifier != null && getMat != null) {
								chunk.foreground[x, y].modifier = getModifier
							}

							val modifierHueShift = reader.readUnsignedByte()

							chunk.foreground[x, y].setModifierHueShift(modifierHueShift)

							val materialID2 = reader.readUnsignedShort()
							val getMat2 = Starbound.tilesAccessID[materialID2]

							if (getMat2 != null) {
								chunk.background[x, y].material = getMat2
								hitTile = true
							}

							// reader.skipBytes(1) // Background hue shift
							// reader.skipBytes(1) // Background color variant

							val colorShift2 = reader.readUnsignedByte()
							val colorVariant2 = reader.readUnsignedByte()

							val modifier2 = reader.readUnsignedShort()
							val getModifier2 = Starbound.tileModifiersByIDAccess[modifier2]

							if (getModifier2 != null && getMat2 != null) {
								chunk.background[x, y].modifier = getModifier2
							}

							chunk.background[x, y].color = colorVariant2
							chunk.background[x, y].setHueShift(colorShift2)

							val modifierHueShift2 = reader.readUnsignedByte()

							chunk.background[x, y].setModifierHueShift(modifierHueShift2)

							val liquid = reader.readUnsignedByte()
							val liquidLevel = reader.readFloat()
							val liquidPressure = reader.readFloat()
							val liquidIsInfinite = reader.readBoolean()
							val collisionMap = reader.readUnsignedByte()
							val dungeonId = reader.readUnsignedShort()
							val biome = reader.readUnsignedByte()
							val envBiome = reader.readUnsignedByte()
							val indestructible = reader.readBoolean()
							val unknown = reader.readUnsignedByte()

							val getLiquid = Starbound.liquidByIDAccess[liquid]

							if (getLiquid != null) {
								val state = chunk.setLiquid(x, y, getLiquid)!!

								state.isInfinite = liquidIsInfinite
								state.pressure = liquidPressure
								state.level = liquidLevel
							}
						}
					}

					set += System.currentTimeMillis() - t

					if (hitTile) {
						//println(chunk.chunk.posVector2d)
						// ent.position = chunk.chunk.posVector2d + Vector2d(16.0, 34.0)
					}
				}
			}
		}

		println("$find $set $parse")

		//client.world!!.parallax = Starbound.parallaxAccess["garden"]
	}

	//ent.position += Vector2d(y = 14.0, x = -10.0)
	ent.position = Vector2d(600.0 + 16.0, 721.0 + 48.0)
	//client.camera.pos.x = 578f
	//client.camera.pos.y = 695f

	client.onDrawGUI {
		client.gl.font.render("${ent.position}", y = 100f, scale = 0.25f)
		client.gl.font.render("${ent.movement.velocity}", y = 120f, scale = 0.25f)
		client.gl.font.render("${client.camera.pos} ${client.settings.zoom}", y = 140f, scale = 0.25f)
	}

	val lightmap = GLFrameBuffer(client.gl)

	lightmap.reattachTexture(client.viewportWidth, client.viewportHeight, GL_RGBA)

	client.onViewportChanged { width, height -> lightmap.reattachTexture(width, height, GL_RGBA) }

	client.onPostDrawWorld {
		lightmap.bind()
		GL46.glClear(GL46.GL_COLOR_BUFFER_BIT or GL46.GL_DEPTH_BUFFER_BIT)
		//client.gl.programs.colorQuad.clearAlpha()

		client.gl.programs.light.use()
		client.gl.programs.light.transform.set(client.gl.matrixStack.last)
		// client.gl.programs.light.transform.set(Matrix4f.IDENTITY)
		client.gl.programs.light.baselineColor.set(Color.WHITE)

		var builder = client.gl.programs.light.builder

		val (rX, rY) = client.screenToWorld(client.mouseCoordinatesF)

		builder.begin()

		//val size = 20f / client.settings.scale
		val size = 10f

		builder.quad(rX - size, rY - size, rX + size, rY + size, QuadTransformers.uv())

		builder.upload()

		//client.gl.blendFunc = BlendFunc(
		//	BlendFunc.Func.DST_ALPHA,
		//	BlendFunc.Func.ONE_MINUS_DST_ALPHA
		//)

		builder.draw()
		client.gl.blendFunc = BlendFunc.ALPHA_TEST

		client.gl.programs.lightOccluder.use()
		client.gl.programs.lightOccluder.transform.set(client.gl.matrixStack.last)

		builder = client.gl.programs.lightOccluder.builder

		builder.begin()

		val lightPosition = client.screenToWorld(client.mouseCoordinatesF)

		builder.quad(0f, 0f, 2f, 2f)
		builder.quad(-6f, 0f, -2f, 2f)

		client.gl.programs.lightOccluder.lightPosition.set(lightPosition)

		builder.upload()

		client.gl.blendFunc = BlendFunc.ONLY_ALPHA
		builder.draw(GL_LINES)
		client.gl.blendFunc = BlendFunc.ALPHA_TEST

		lightmap.unbind()

		client.gl.activeTexture = 0
		client.gl.texture2D = lightmap.texture
		client.gl.programs.textureQuad.run(0)

		client.gl.programs.flat.use()
		client.gl.programs.flat.color.set(Color.BLACK)
		client.gl.programs.flat.transform.set(client.gl.matrixStack.last)
		builder = client.gl.programs.flat.builder

		builder.begin()

		builder.quad(0f, 0f, 2f, 2f)
		builder.quad(-6f, 0f, -2f, 2f)

		builder.upload()
		builder.draw()
	}

	client.onPreDrawWorld {
		//client.camera.pos.x = ent.pos.x.toFloat()
		//client.camera.pos.y = ent.pos.y.toFloat()
	}

	client.gl.box2dRenderer.drawShapes = false
	client.gl.box2dRenderer.drawPairs = false
	client.gl.box2dRenderer.drawAABB = false
	client.gl.box2dRenderer.drawJoints = false

	//ent.spawn()

	client.input.addScrollCallback { _, x, y ->
		if (y > 0.0) {
			client.settings.zoom *= y.toFloat() * 2f
		} else if (y < 0.0) {
			client.settings.zoom /= -y.toFloat() * 2f
		}
	}

	while (client.renderFrame()) {
		Starbound.pollCallbacks()

		//ent.think(client.frameRenderTime)
		//client.camera.pos.x = ent.position.x.toFloat()
		//client.camera.pos.y = ent.position.y.toFloat()

		client.camera.pos.x += if (client.input.KEY_LEFT_DOWN || client.input.KEY_A_DOWN) -client.frameRenderTime.toFloat() * 32f / client.settings.zoom else 0f
		client.camera.pos.x += if (client.input.KEY_RIGHT_DOWN || client.input.KEY_D_DOWN) client.frameRenderTime.toFloat() * 32f / client.settings.zoom else 0f

		client.camera.pos.y += if (client.input.KEY_UP_DOWN || client.input.KEY_W_DOWN) client.frameRenderTime.toFloat() * 32f / client.settings.zoom else 0f
		client.camera.pos.y += if (client.input.KEY_DOWN_DOWN || client.input.KEY_S_DOWN) -client.frameRenderTime.toFloat() * 32f / client.settings.zoom else 0f

		//println(client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1)

		//if (ent.onGround)
			//ent.velocity += client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1

		/*if (client.input.KEY_LEFT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_LEFT
		} else if (client.input.KEY_RIGHT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_RIGHT
		} else {
			ent.movement.moveDirection = Move.STAND_STILL
		}

		if (client.input.KEY_SPACE_PRESSED && ent.movement.onGround) {
			ent.movement.requestJump()
		} else if (client.input.KEY_SPACE_RELEASED) {
			ent.movement.recallJump()
		}*/

		if (client.input.KEY_ESCAPE_PRESSED) {
			glfwSetWindowShouldClose(client.window, true)
		}

		//ent.wantsToDuck = client.input.KEY_DOWN_DOWN

		//if (chunkA != null && glfwGetTime() < 10.0) {
		//	val tile = Starbound.getTileDefinition("alienrock")
			//chunkA!!.foreground[rand.nextInt(0, CHUNK_SIZE_FF), rand.nextInt(0, CHUNK_SIZE_FF)] = tile
		//}
	}
}
